@@ -2,6 +2,7 @@ from tests import unittest
 
 from botocore import model
 from botocore.compat import OrderedDict
+from botocore.exceptions import MissingServiceIdError
 
 
 def test_missing_model_attribute_raises_exception():
@@ -30,18 +31,44 @@ def test_missing_model_attribute_raises_exception():
         yield _test_attribute_raise_exception, name
 
 
+class TestServiceId(unittest.TestCase):
+    def test_hypenize_replaces_spaces(self):
+        self.assertEqual(
+            model.ServiceId('my service').hyphenize(), 'my-service'
+        )
+
+    def test_hyphenize_lower_cases(self):
+        self.assertEqual(model.ServiceId('MyService').hyphenize(), 'myservice')
+
+
 class TestServiceModel(unittest.TestCase):
 
     def setUp(self):
         self.model = {
             'metadata': {'protocol': 'query',
-                         'endpointPrefix': 'endpoint-prefix'},
+                         'endpointPrefix': 'endpoint-prefix',
+                         'serviceId': 'MyService'},
             'documentation': 'Documentation value',
             'operations': {},
             'shapes': {
                 'StringShape': {'type': 'string'}
             }
         }
+        self.error_shapes = {
+            'ExceptionOne': {
+                'exception': True,
+                'type': 'structure',
+                'members': {},
+            },
+            'ExceptionTwo': {
+                'exception': True,
+                'type': 'structure',
+                'members': {},
+                'error': {
+                    'code': 'FooCode'
+                }
+            },
+        }
         self.service_model = model.ServiceModel(self.model)
 
     def test_metadata_available(self):
@@ -57,6 +84,31 @@ class TestServiceModel(unittest.TestCase):
     def test_service_name_defaults_to_endpoint_prefix(self):
         self.assertEqual(self.service_model.service_name, 'endpoint-prefix')
 
+    def test_service_id(self):
+        self.assertEqual(self.service_model.service_id, 'MyService')
+
+    def test_hyphenize_service_id(self):
+        self.assertEqual(
+            self.service_model.service_id.hyphenize(), 'myservice')
+
+    def test_service_id_does_not_exist(self):
+        service_model = {
+            'metadata': {
+                'protocol': 'query',
+                'endpointPrefix': 'endpoint-prefix',
+            },
+            'documentation': 'Documentation value',
+            'operations': {},
+            'shapes': {
+                'StringShape': {'type': 'string'}
+            }
+        }
+        service_name = 'myservice'
+        service_model = model.ServiceModel(service_model, service_name)
+        with self.assertRaisesRegexp(model.UndefinedModelAttributeError,
+                                     service_name):
+            service_model.service_id
+
     def test_operation_does_not_exist(self):
         with self.assertRaises(model.OperationNotFoundError):
             self.service_model.operation_model('NoExistOperation')
@@ -71,6 +123,27 @@ class TestServiceModel(unittest.TestCase):
     def test_shape_names(self):
         self.assertEqual(self.service_model.shape_names, ['StringShape'])
 
+    def test_repr_has_service_name(self):
+        self.assertEqual(repr(self.service_model),
+                         'ServiceModel(endpoint-prefix)')
+
+    def test_shape_for_error_code(self):
+        self.model['shapes'].update(self.error_shapes)
+        self.service_model = model.ServiceModel(self.model)
+        shape = self.service_model.shape_for_error_code('ExceptionOne')
+        self.assertEqual(shape.name, 'ExceptionOne')
+        shape = self.service_model.shape_for_error_code('FooCode')
+        self.assertEqual(shape.name, 'ExceptionTwo')
+
+    def test_error_shapes(self):
+        self.model['shapes'].update(self.error_shapes)
+        self.service_model = model.ServiceModel(self.model)
+        error_shapes = self.service_model.error_shapes
+        error_shape_names = [shape.name for shape in error_shapes]
+        self.assertEqual(len(error_shape_names), 2)
+        self.assertIn('ExceptionOne', error_shape_names)
+        self.assertIn('ExceptionTwo', error_shape_names)
+
 
 class TestOperationModelFromService(unittest.TestCase):
     def setUp(self):
@@ -242,6 +315,148 @@ class TestOperationModelFromService(unittest.TestCase):
         operation_two = service_model.operation_model('OperationTwo')
         self.assertFalse(operation_two.deprecated)
 
+    def test_endpoint_operation_present(self):
+        self.model['operations']['OperationName']['endpointoperation'] = True
+        service_model = model.ServiceModel(self.model)
+        operation_name = service_model.operation_model('OperationName')
+        self.assertTrue(operation_name.is_endpoint_discovery_operation)
+
+    def test_endpoint_operation_present_false(self):
+        self.model['operations']['OperationName']['endpointoperation'] = False
+        service_model = model.ServiceModel(self.model)
+        operation_name = service_model.operation_model('OperationName')
+        self.assertFalse(operation_name.is_endpoint_discovery_operation)
+
+    def test_endpoint_operation_absent(self):
+        operation_two = self.service_model.operation_model('OperationName')
+        self.assertFalse(operation_two.is_endpoint_discovery_operation)
+
+    def test_endpoint_discovery_present(self):
+        operation = self.model['operations']['OperationName']
+        operation['endpointdiscovery'] = {'required': True}
+        service_model = model.ServiceModel(self.model)
+        operation_name = service_model.operation_model('OperationName')
+        self.assertTrue(operation_name.endpoint_discovery.get('required'))
+
+    def test_endpoint_discovery_absent(self):
+        operation_name = self.service_model.operation_model('OperationName')
+        self.assertIsNone(operation_name.endpoint_discovery)
+
+
+class TestOperationModelEventStreamTypes(unittest.TestCase):
+    def setUp(self):
+        super(TestOperationModelEventStreamTypes, self).setUp()
+        self.model = {
+            'metadata': {'protocol': 'rest-xml', 'endpointPrefix': 'foo'},
+            'documentation': '',
+            'operations': {
+                'OperationName': {
+                    'http': {
+                        'method': 'POST',
+                        'requestUri': '/',
+                    },
+                    'name': 'OperationName',
+                    'input': {'shape': 'OperationRequest'},
+                    'output': {'shape': 'OperationResponse'},
+                }
+            },
+            'shapes': {
+                'NormalStructure': {
+                    'type': 'structure',
+                    'members': {
+                        'Input': {'shape': 'StringType'}
+                    }
+                },
+                'OperationRequest': {
+                    'type': 'structure',
+                    'members': {
+                        'String': {'shape': 'StringType'},
+                        "Body": {'shape': 'EventStreamStructure'}
+                    },
+                    'payload': 'Body'
+                },
+                'OperationResponse': {
+                    'type': 'structure',
+                    'members': {
+                        'String': {'shape': 'StringType'},
+                        "Body": {'shape': 'EventStreamStructure'}
+                    },
+                    'payload': 'Body'
+                },
+                'StringType': {'type': 'string'},
+                'BlobType': {'type': 'blob'},
+                'EventStreamStructure': {
+                    'eventstream': True,
+                    'type': 'structure',
+                    'members': {
+                        'EventA': {'shape': 'EventAStructure'},
+                        'EventB': {'shape': 'EventBStructure'}
+                    }
+                },
+                'EventAStructure': {
+                    'event': True,
+                    'type': 'structure',
+                    'members': {
+                        'Payload': {
+                            'shape': 'BlobType',
+                            'eventpayload': True
+                        },
+                        'Header': {
+                            'shape': 'StringType',
+                            'eventheader': True
+                        }
+                    }
+                },
+                'EventBStructure': {
+                    'event': True,
+                    'type': 'structure',
+                    'members': {
+                        'Records': {'shape': 'StringType'}
+                    }
+                }
+            }
+        }
+
+    def update_operation(self, **kwargs):
+        operation = self.model['operations']['OperationName']
+        operation.update(kwargs)
+
+    def test_event_stream_input_for_operation(self):
+        service_model = model.ServiceModel(self.model)
+        operation = service_model.operation_model('OperationName')
+        self.assertTrue(operation.has_event_stream_input)
+        event_stream_input = operation.get_event_stream_input()
+        self.assertEqual(event_stream_input.name, 'EventStreamStructure')
+
+    def test_no_event_stream_input_for_operation(self):
+        self.update_operation(input={'shape': 'NormalStructure'})
+        service_model = model.ServiceModel(self.model)
+        operation = service_model.operation_model('OperationName')
+        self.assertFalse(operation.has_event_stream_input)
+        self.assertEqual(operation.get_event_stream_input(), None)
+
+    def test_event_stream_output_for_operation(self):
+        service_model = model.ServiceModel(self.model)
+        operation = service_model.operation_model('OperationName')
+        self.assertTrue(operation.has_event_stream_output)
+        output = operation.get_event_stream_output()
+        self.assertEqual(output.name, 'EventStreamStructure')
+
+    def test_no_event_stream_output_for_operation(self):
+        self.update_operation(output={'shape': 'NormalStructure'})
+        service_model = model.ServiceModel(self.model)
+        operation = service_model.operation_model('OperationName')
+        self.assertFalse(operation.has_event_stream_output)
+        self.assertEqual(operation.get_event_stream_output(), None)
+
+    def test_no_output_shape(self):
+        self.update_operation(output=None)
+        del self.model['operations']['OperationName']['output']
+        service_model = model.ServiceModel(self.model)
+        operation = service_model.operation_model('OperationName')
+        self.assertFalse(operation.has_event_stream_output)
+        self.assertEqual(operation.get_event_stream_output(), None)
+
 
 class TestOperationModelStreamingTypes(unittest.TestCase):
     def setUp(self):
@@ -492,6 +707,27 @@ class TestShapeResolver(unittest.TestCase):
                          ['NewPassword', 'OldPassword'])
         self.assertEqual(shape.members['OldPassword'].name, 'passwordType')
         self.assertEqual(shape.members['OldPassword'].type_name, 'string')
+        self.assertEqual(shape.error_code, None)
+
+    def test_exception_error_code(self):
+        shapes = {
+            'FooException': {
+                'exception': True,
+                'type': 'structure',
+                'members': {}
+            }
+        }
+        # Test without explicit error code
+        resolver = model.ShapeResolver(shapes)
+        shape = resolver.get_shape_by_name('FooException')
+        self.assertTrue(shape.metadata['exception'])
+        self.assertEqual(shape.error_code, 'FooException')
+        # Test with explicit error code
+        shapes['FooException']['error'] = {'code': 'ExceptionCode'}
+        resolver = model.ShapeResolver(shapes)
+        shape = resolver.get_shape_by_name('FooException')
+        self.assertTrue(shape.metadata['exception'])
+        self.assertEqual(shape.error_code, 'ExceptionCode')
 
     def test_shape_metadata(self):
         shapes = {
@@ -519,6 +755,26 @@ class TestShapeResolver(unittest.TestCase):
         self.assertEqual(member.metadata['max'], 128)
         self.assertEqual(member.metadata['sensitive'], True)
 
+    def test_error_shape_metadata(self):
+        shapes = {
+            'ResourceNotFoundException': {
+                'type': 'structure',
+                'members': {
+                    'message': {
+                        'shape': 'ErrorMessage',
+                    }
+                },
+                'exception': True,
+                'retryable': {'throttling': True}
+            }
+        }
+        resolver = model.ShapeResolver(shapes)
+        shape = resolver.get_shape_by_name('ResourceNotFoundException')
+        self.assertEqual(
+            shape.metadata,
+            {'exception': True, 'retryable': {'throttling': True}}
+        )
+
     def test_shape_list(self):
         shapes = {
             'mfaDeviceListType': {
@@ -687,6 +943,20 @@ class TestBuilders(unittest.TestCase):
         self.assertEqual(shape.members['A'].documentation,
                          'MyDocs')
 
+    def test_min_max_used_in_metadata(self):
+        b = model.DenormalizedStructureBuilder()
+        shape = b.with_members({
+            'A': {
+                'type': 'string',
+                'documentation': 'MyDocs',
+                'min': 2,
+                'max': 3,
+            },
+        }).build_model()
+        metadata = shape.members['A'].metadata
+        self.assertEqual(metadata.get('min'), 2)
+        self.assertEqual(metadata.get('max'), 3)
+
     def test_use_shape_name_when_provided(self):
         b = model.DenormalizedStructureBuilder()
         shape = b.with_members({
